    
import math
import random
import parmap
import itertools
from itertools import repeat
from scipy.spatial import distance
import operator
from tqdm import tqdm
from functools import reduce
from sklearn import mixture
import statsmodels.stats.multitest as multi
import networkx as nx
import multiprocessing as mp
import numpy as np
from scipy.stats import poisson
import pandas as pd
import pygco as pygco # cut_from_graph # pip install git+git://github.com/amueller/gco_python
import scipy.stats.mstats as ms
from itertools import repeat
from scipy.stats import norm
from scipy.stats import binom
from scipy.stats import skewnorm
from scipy.sparse import issparse
import matplotlib as mpl
import matplotlib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.path as mplPath
import shapely.geometry
import shapely.ops
from scipy.spatial import Voronoi, voronoi_plot_2d, Delaunay, KDTree, ConvexHull
from matplotlib.patches import Polygon
from matplotlib.collections import LineCollection, PatchCollection
from PIL import Image
from matplotlib.backends.backend_pdf import PdfPages
import sklearn.manifold as manifold
import sklearn.decomposition as decomposition 
from sklearn.preprocessing import StandardScaler
from scipy.stats import gaussian_kde
from sklearn.cluster import DBSCAN,KMeans
from scipy.stats import ttest_ind
from sklearn.utils import shuffle
import hdbscan
import seaborn as sns
import sklearn.cluster as cluster
from sklearn.cluster import KMeans
from scipy.spatial.distance import cdist
from scipy import stats
import sys
import os
import time
from scipy.stats import skewnorm
from scipy import stats
from ast import literal_eval
import pysal
from pysal.explore.esda.moran import Moran
from skimage.filters import threshold_otsu

from .GMM import *


def create_graph_with_weight(points, normCount):
    '''
    Returns a graph created from cell coordiantes.
    edge weights set by normalized counts.
    
    :param points: shape (n,2); normCount: shape (n)
    :rtype: ndarray shape (n ,3)
    
    '''
    edges = {}   
    var = normCount.var()
    delauny = Delaunay(points)
    
## 1. get original all connected edges and remove repetitives edges among triangle
    cellGraph = np.zeros((points.shape[0]*10, 4))
    for simplex in delauny.simplices:
        simplex.sort()
        edge0 = str(simplex[0]) + " " + str(simplex[1])
        edge1 = str(simplex[0]) + " " + str(simplex[2])
        edge2 = str(simplex[1]) + " " + str(simplex[2])
        edges[edge0] = 1
        edges[edge1] = 1
        edges[edge2] = 1
        
    i = 0
    for kk in edges.keys():  
        node0 = int(kk.split(sep=" ")[0])
        node1 = int(kk.split(sep=" ")[1])
        edgeDiff = normCount[node0] - normCount[node1]
        energy = np.exp((0 - edgeDiff**2)/(2*var))
        dist = distance.euclidean(points[node0,:], points[node1,:])
        cellGraph[i] = [node0, node1, energy, dist]       
        i = i + 1
    tempGraph = cellGraph[0:i]
    
## 2. remove more than cutoff edges from longest edges of each simplex taiangle. 
    n_components_range = range(1,5)
    best_component = 1
    lowest_bic=np.infty
    temp_data = tempGraph[:,3].reshape(-1,1)  ## GMM of dist 
    for n_components in n_components_range:
        gmm = mixture.GaussianMixture(n_components = n_components, random_state=2020)
        gmm.fit(temp_data)
        gmm_bic = gmm.bic(temp_data)
        if gmm_bic < lowest_bic:
            best_gmm = gmm
            lowest_bic = gmm_bic
            best_component = n_components  
    
    mIndex = np.where(best_gmm.weights_ == max(best_gmm.weights_))[0]
    cutoff = best_gmm.means_[mIndex] + 2*np.sqrt(best_gmm.covariances_[mIndex])

    for simplex in delauny.simplices:
        simplex.sort()          
        dist0 = distance.euclidean(points[simplex[0],:], points[simplex[1],:])
        dist1 = distance.euclidean(points[simplex[0],:], points[simplex[2],:])
        dist2 = distance.euclidean(points[simplex[1],:], points[simplex[2],:])
        tempArray = np.array((dist0, dist1, dist2))
        badIndex = np.where(tempArray == max(tempArray))[0][0]  ## remove longest edges among simplex taiangle.
        if tempArray[badIndex] > cutoff:
            edge0 = str(simplex[0]) + " " + str(simplex[1])  
            edge1 = str(simplex[0]) + " " + str(simplex[2])       
            edge2 = str(simplex[1]) + " " + str(simplex[2])
            edgeCount = 0
            if edge0 in edges and edge1 in edges and edge2 in edges:
                if badIndex == 0:
                    del edges[edge0]
                elif badIndex == 1:
                    del edges[edge1]
                elif badIndex == 2:
                    del edges[edge2]     ## remove longest edges from edges

    i = 0
    for kk in edges.keys():         ## recrete cellGraph with new edges
        node0 = int(kk.split(sep=" ")[0])
        node1 = int(kk.split(sep=" ")[1])
        edgeDiff = normCount[node0] - normCount[node1]
        energy = np.exp((0 - edgeDiff**2)/(2*var))
        dist = distance.euclidean(points[node0,:], points[node1,:])
        cellGraph[i] = [node0, node1, energy, dist]       
        i = i + 1   
    tempGraph = cellGraph[0:i] # tempGraph: no need remove these longest edges among all edges
    
# 3.  test these longest edges among global distance 
    temp_data = tempGraph[:,3].reshape(-1,1)    
    gmm = mixture.GaussianMixture(n_components = 1,random_state=2020)
    gmm.fit(temp_data)    
    cutoff = gmm.means_[0] + 2*np.sqrt(gmm.covariances_[0])
    tempGraph1 = tempGraph.copy()  
    j=0
    for i in np.arange(tempGraph.shape[0]):    
        if tempGraph[i, 3] < cutoff:     ### 
            tempGraph1[j] = tempGraph[i] # tempGraph2: need remove these longest edges among all edges
            j = j + 1
    
# 4. count broken edges  20211030 by coco
    tempGraph2 = tempGraph1[0:j].copy()
    G_full = nx.Graph()
    G_full.add_nodes_from(set(np.arange(points.shape[0])))
    G_full.add_edges_from(tempGraph2[:, 0:2].astype(np.int32))
    # isolated_nodes = list(nx.isolates(G_full))
    # len(isolated_nodes) 
    deg_dict = {key: value for (key, value) in list(G_full.degree())}
    Nbroken= sum(np.array(list(deg_dict.values()))<2)
    if Nbroken >3:
        finalGraph = tempGraph  # too many broken edges, no need remove these longest edges among all edges
    else:
        finalGraph = tempGraph1  # tempGraph2: need remove these longest edges among all edges
    return finalGraph


def first_neg_index(a):
    '''
    deprecated
    '''
    for i in np.arange(a.shape[0]):
        if a[i] < 0:
            return i
    return a.shape[0] - 1                

def calc_u_cost(a, mid_points):
    '''
    deprecated
    '''
    neg_index = int(a[0])
    x = a[1]
    m_arr = np.concatenate((0 - mid_points[0:neg_index+1], 
                            mid_points[neg_index:]), axis=0)
    x_arr = np.concatenate((np.repeat(x, neg_index+1), 
                0 - np.repeat(x, mid_points.shape[0] - neg_index)), axis=0)
    return m_arr+x_arr   


def compute_pairwise_cost(size, smooth_factor):
    '''
    Returns pairwise energy.
    
    :param points: size: scalar; smooth_factor: scalar

    :rtype: pairwise energy matrix.
    '''
    pairwise_size = size
    pairwise = -smooth_factor * np.eye(pairwise_size, dtype=np.int32)
    step_weight = -smooth_factor*np.arange(pairwise_size)[::-1]
    for i in range(pairwise_size): 
        pairwise[i,:] += np.roll(step_weight,i) 
    temp = np.triu(pairwise).T + np.triu(pairwise)
    np.fill_diagonal(temp, np.diag(temp)/2)
    return temp

def cut_graph_general_profile(cellGraph, count,gmm, unary_scale_factor=100, 
                      smooth_factor=50, label_cost=10, algorithm='expansion'):
    '''
    Returns new labels and gmm for the cut with gmm profile.
    
    :param points: cellGraph (n,3); count: shape (n,); 
    :unary_scale_factor, scalar; smooth_factor, scalar; 
    :label_cost: scalar; algorithm='expansion'
    :rtype: label shape (n,); gmm object.
    '''
    unary_scale_factor = unary_scale_factor
    label_cost = label_cost
    algorithm = algorithm
    smooth_factor = smooth_factor
    gmm=gmm      
    unary_cost = compute_unary_cost_simple_profile(count, gmm, unary_scale_factor)
    
    pairwise_cost = compute_pairwise_cost(gmm.means_.shape[0], smooth_factor)
    edges = cellGraph[:,0:2].astype(np.int32)
    labels = pygco.cut_from_graph(edges, unary_cost, pairwise_cost, label_cost)
#    energy = compute_energy(unary_cost, pairwise_cost, edges, labels)

    return labels

def compute_unary_cost_simple_profile(count, gmm, unary_scale_factor):
    '''
    Returns unary cost energy.
    
    :param points: count: shape (n,); gmm: gmm object; scale_factor: scalar

    :rtype: unary energy matrix.
    '''    
    exp = count
    a= exp[exp > 0]
    gmm_pred = gmm.predict(a.reshape(-1,1))
    zero_label = gmm.predict(np.min(a).reshape(-1,1))[0]
    label_pred = gmm.predict(exp.reshape(-1,1))
    if len(np.where(exp == 0)[0]) > 0:
        np.place(label_pred, exp==0, zero_label)
    uniq, count = np.unique(label_pred, return_counts = True)    

    if len(uniq)<2:  ## when just one uniq,the will mismatch dimension with pairwaise label
        uniq_modify =np.append(uniq,uniq)
    else:
        uniq_modify =uniq
    uninary_mat = np.zeros((len(label_pred), len(uniq_modify)))
    for i in np.arange(uninary_mat.shape[0]):
        for j in np.arange(len(uniq_modify)):
            if uniq_modify[j] == label_pred[i]:  ## same ,energy -1; imsame ,energy 1.
                uninary_mat[i, j] = -1
            else:
                uninary_mat[i, j] = 1   
    return (unary_scale_factor*uninary_mat).astype(np.int32), label_pred


def cut_graph_general(cellGraph, count,gmm, unary_scale_factor=100, 
                      smooth_factor=30, label_cost=10, algorithm='expansion',
                      profile=False):
    '''
    Returns new labels and gmm for the cut.
    
    :param points: cellGraph (n,3); count: shape (n,); 
    :unary_scale_factor, scalar; smooth_factor, scalar; 
    :label_cost: scalar; algorithm='expansion'
    :rtype: label shape (n,); gmm object.
    '''
    unary_scale_factor = unary_scale_factor
    label_cost = label_cost
    algorithm = algorithm
    smooth_factor = smooth_factor
    gmm=gmm
#     a = count.copy() 
#     if sum(a>1)/len(count) <= 0.1 and sum(a>1)<=30:             ###  1-1, 0.25 can modify 
#         np.place(a, a==0, (np.random.rand(sum(a==0))*0.25)) # using 0.1 gives layer 4 300; 0.25 gives 150; 0.5 gives 80?
#         gmm = find_mixture_2(a)       
#     else:
#         a=a[a>0]
#         gmm = find_mixture(a)
    
    if profile==False:
        unary_cost, label_pred = compute_unary_cost_simple(count,  gmm,  unary_scale_factor)
    else:
        unary_cost, label_pred = compute_unary_cost_simple_profile(count, gmm, unary_scale_factor)
    
    pairwise_cost = compute_pairwise_cost(gmm.means_.shape[0], smooth_factor)
    edges = cellGraph[:,0:2].astype(np.int32)
    labels = pygco.cut_from_graph(edges, unary_cost,
                                 pairwise_cost.astype(np.int32), np.int32(label_cost))
#    energy = compute_energy(unary_cost, pairwise_cost, edges, labels)

    return labels, label_pred



def compute_unary_cost_simple(count, gmm, scale_factor):
    '''
    Returns unary cost energy.
    
    :param points: count: shape (n,); gmm: gmm object; scale_factor: scalar

    :rtype: unary energy matrix.
    '''    
    exp = count
    a= exp[exp > 0]

    gmm_pred = gmm.predict(a.reshape(-1,1))
    zero_label = gmm.predict(np.min(a).reshape(-1,1))[0]
    label_pred = gmm.predict(exp.reshape(-1,1))
    if len(np.where(exp == 0)[0]) > 0:
        np.place(label_pred, exp==0, zero_label)
    
    temp_means = np.sort(gmm.means_, axis=None)
   # new_index = np.where(gmm.means_ == temp_means)[1]
    # temp_covs = gmm.covariances_.copy()
    # for i in np.arange(new_index.shape[0]):
    #     temp_covs[i] = gmm.covariances_[new_index[i]]  ##sorted gmm.covariances as np.sort(gmm.means_)
    new_index= np.argsort(gmm.means_,axis=0).reshape(1,-1)[0]
    temp_covs = gmm.covariances_[new_index]

    new_labels = np.zeros(label_pred.shape[0], dtype=np.int32)
    for i in np.arange(new_index.shape[0]):
        temp_index = np.where(label_pred == i)[0]
        new_labels[temp_index] = new_index[i]  ## the label of min(gmm.means_) is 0; label of max is 1.

    mid_points = np.zeros(len(new_index) - 1)
    for i in np.arange(len(mid_points)):
        mid_points[i] = (temp_means[i]*np.sqrt(temp_covs[i+1]) + 
                     temp_means[i+1]*np.sqrt(temp_covs[i])
                    )/(np.sqrt(temp_covs[i]) + np.sqrt(temp_covs[i+1]))
    temp = count[:, np.newaxis] - temp_means.T[1:]
    neg_indices = np.apply_along_axis(first_neg_index, 1, temp)
    ind_count_arr = np.vstack((neg_indices, count)).T        
    unary_cost =  (scale_factor*np.apply_along_axis(calc_u_cost, 1, 
                                    ind_count_arr, mid_points)).astype(np.int32)
    a = count.copy()
    if sum(a>1)/len(count) <= 0.1 and sum(a>1)<=30:
        unique_cost, counts_cost = np.unique(unary_cost, return_counts=True)
        val_to_replace1 = unique_cost[np.argmax(counts_cost)]
        val_to_replace2 = 0 - unique_cost[np.argmax(counts_cost)]
        if val_to_replace1 < 10 and val_to_replace1 > 0:
            np.place(unary_cost, unary_cost == val_to_replace1, np.median(abs(unique_cost)))
        elif val_to_replace1 < 0 and val_to_replace1 > -10:
            np.place(unary_cost, unary_cost == val_to_replace1, 0-np.median(abs(unique_cost)))
        if val_to_replace2 < 10 and val_to_replace2 > 0:
            np.place(unary_cost, unary_cost == val_to_replace2, np.median(abs(unique_cost)))
        elif val_to_replace2 < 0 and val_to_replace2 > -10:
            np.place(unary_cost, unary_cost == val_to_replace2, 0-np.median(abs(unique_cost)))
    return unary_cost, label_pred


def noise_inside(a, b, cellGraph):
    idx0 = np.in1d(cellGraph[:,0], np.array(list(a))).nonzero()[0]
    idx1 = np.in1d(cellGraph[:,1], np.array(list(a))).nonzero()[0]
    neighbor0 = cellGraph[idx0, 1]
    neighbor1 = cellGraph[idx1, 0]
    neighbors = set(neighbor0.tolist() + neighbor1.tolist())   
    out_neighbors = neighbors.difference(set(a))
    not_a_neighbors = out_neighbors.difference(set(b))
    return (len(not_a_neighbors) == 0)


# added noise
def compute_p_CSR(locs, newLabels, gmm, exp, cellGraph): 
    '''
    Returns p_value of the cut.
    
    :param points: newLabels: shape (n,); gmm: gmm object
                   exp: ndarray shape (n ,3); cellGraph: shape (n,3)

    :rtype: p_value.
    '''
    
    p_values = list()
    node_lists = list()
    gmm_pred=gmm_predict(exp,gmm)
    unique, counts = np.unique(gmm_pred,return_counts=True)
    con_components = count_component(locs,cellGraph, newLabels)  ## nodes index in subgraphs
    noise = dict()
    # now calculate p for all comp without considering noise

    min_sig_p_size = np.inf
    for j in np.arange(len(con_components)):
        node_list = con_components[j]
        com_size = len(node_list)
        if len(exp)<= 1000:
            com_cutoff = 3
            addNoise = True
        else:
            com_cutoff = 9
            addNoise = False
        if com_size >= com_cutoff:
                # we want to score the object not the back ground
            gmm_pred_com=gmm_pred[list(node_list)]          
            # check 0s
            unique_com, counts_com = np.unique(gmm_pred_com, return_counts=True)  
            major_label = unique_com[np.where(counts_com == counts_com.max())[0][0]]
            label_count = counts[np.where(unique == major_label)[0]]  ##  real counts that in subgraphs,Ci
            count_in_com =  counts_com.max()   ## counts by graph cuts,k
            cover = exp.shape[0]/com_size
            p0 = poisson.sf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
            p1 = poisson.pmf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
            prob=min((p0+p1),1)  
            if prob < 10e-30:
                prob = 10e-30
            p_values.append(prob)
            if prob <0.1 and len(con_components[j]) < min_sig_p_size:
                min_sig_p_size = len(con_components[j])
        else: # set small comp p=1
            p_values.append(1)            
        node_lists.append(np.array(list(node_list)))
    
    if addNoise:
        for j in np.arange(len(con_components)):
            if p_values[j] >= 0.1 and len(con_components[j]) < 10: # min_sig_p_size:
                noise[j] = con_components[j]
    # now re-calculate p considering noise
        for j in np.arange(len(con_components)):
            if p_values[j] < 0.1: # small p let correct by consider noise
                noise_size = 0
                used_com = list()
                for jj, cc in noise.items():
                    if noise_inside(cc, con_components[j], cellGraph):
                        noise_size = noise_size + len(cc)
                        used_com.append(jj)
                if noise_size > 0:
                    for jj in used_com:
                        noise.pop(jj)
                    node_list = con_components[j]
                    com_size = len(node_list)
                    gmm_pred_com=gmm_pred[list(node_list)]
                    unique_com, counts_com = np.unique(gmm_pred_com, return_counts=True)  
                    major_label = unique_com[np.where(counts_com == counts_com.max())[0][0]]
                    label_count = counts[np.where(unique == major_label)[0]]  ##  real counts that in subgraphs,Ci
                    count_in_com =  counts_com.max()   ## counts by graph cuts,k
                    com_size = com_size + noise_size
                    cover = exp.shape[0]/com_size
                
                    p0 = poisson.sf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
                    p1 = poisson.pmf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
                    prob=min((p0+p1),1)
                    if prob < 10e-30:
                        prob = 10e-30  
                    p_values[j] = prob               
    return p_values, node_lists, con_components


def count_component(locs, cellGraph, newLabels):
    '''
    Returns number of subgraphs.
    
    :param points: cellGraph: shape (n,3); newLabels: ndarray shape (n,); locs: shape (n, 2) 

    :rtype: scalar. 
    
    '''

    G_cut = nx.Graph()
    tempGraph = cellGraph.copy()

    tempGraph = np.apply_along_axis(remove_egdes, 1, tempGraph, newLabels)   ## reassign the edges energy between two nodes,save cellGraph[:,2]
    G_cut.add_nodes_from(list(set(list(tempGraph[:,0].astype(np.int32)) + list(tempGraph[:,1].astype(np.int32)))))    
    G_cut.add_edges_from(tempGraph[np.where(tempGraph[:,2] == 1)[0],0:2].astype(np.int32))  ## connect same label to one subgraph.
    
    com = sorted(nx.connected_components(G_cut),    ## sort by len(# nodes in subgraphs), com[1] is second largest subgraphs
                                  key = len, reverse=True)  

    return com  

def count_isolate(locs, cellGraph, newLabels):
    '''
    Returns number of subgraphs.
    
    :param points: cellGraph: shape (n,3); newLabels: ndarray shape (n,); locs: shape (n, 2) 

    :rtype: scalar. 
    
    '''

    G_full = nx.Graph()
    G_cut = nx.Graph()
    tempGraph = cellGraph.copy()
    G_full.add_nodes_from(list(set(list(tempGraph[:,0].astype(np.int32)) + list(tempGraph[:,1].astype(np.int32)))))
    G_full.add_edges_from(tempGraph[:, 0:2].astype(np.int32))   
    tempGraph = np.apply_along_axis(remove_egdes, 1, tempGraph, newLabels)
    G_cut.add_nodes_from(list(set(list(tempGraph[:,0].astype(np.int32)) + list(tempGraph[:,1].astype(np.int32)))))    
    G_cut.add_edges_from(tempGraph[np.where(tempGraph[:,2] == 1)[0],0:2].astype(np.int32))
    degree_cutoff = int(min(nx.average_degree_connectivity(G_full).values()))
    com = sorted(nx.connected_components(G_cut), 
                                  key = len, reverse=True)  
    isolated_nodes = list(nx.isolates(G_cut))
    seg_size = np.zeros(18)
    if len(isolated_nodes) > 0:
        deg_dict = {key: value for (key, value) in list(G_full.degree(list(nx.isolates(G_cut))))}
#    seg_size[0] = (np.array(list(deg_dict.values())) > degree_cutoff).sum()             
#    t_com = len(com) + locs.shape[0] - sum_nodes 
#    t_com = locs.shape[0] - sum_nodes 
        seg_size[0] = (np.array(list(deg_dict.values())) <= degree_cutoff).sum() 
        seg_size[1] = (np.array(list(deg_dict.values())) > degree_cutoff).sum() 
    for cc in com:
        if len(cc) == 2:
            deg_dict = {key: value for (key, value) in list(G_full.degree(list(cc)))}
            if (np.array(list(deg_dict.values())) <= degree_cutoff).sum() > 0:
                seg_size[2] = seg_size[2] + 1
            else:
                seg_size[3] = seg_size[3] + 1
                
        if len(cc) == 3:
            deg_dict = {key: value for (key, value) in list(G_full.degree(list(cc)))}
            if (np.array(list(deg_dict.values())) <= degree_cutoff).sum() >= 2:
                seg_size[4] = seg_size[4] + 1
            else:
                seg_size[5] = seg_size[5] + 1  
    
        if len(cc) == 4:
            deg_dict = {key: value for (key, value) in list(G_full.degree(list(cc)))}
            if (np.array(list(deg_dict.values())) <= degree_cutoff).sum() >= 2:
                seg_size[6] = seg_size[6] + 1
            else:
                seg_size[7] = seg_size[7] + 1 
        if len(cc) == 5:
            deg_dict = {key: value for (key, value) in list(G_full.degree(list(cc)))}
            if (np.array(list(deg_dict.values())) <= degree_cutoff).sum() >= 2:
                seg_size[8] = seg_size[8] + 1
            else:
                seg_size[9] = seg_size[9] + 1 
        if len(cc) == 6:
            deg_dict = {key: value for (key, value) in list(G_full.degree(list(cc)))}
            if (np.array(list(deg_dict.values())) <= degree_cutoff).sum() >= 2:
                seg_size[10] = seg_size[10] + 1
            else:
                seg_size[11] = seg_size[11] + 1 
        if len(cc) == 7:
            deg_dict = {key: value for (key, value) in list(G_full.degree(list(cc)))}
            if (np.array(list(deg_dict.values())) <= degree_cutoff).sum() >= 2:
                seg_size[12] = seg_size[12] + 1
            else:
                seg_size[13] = seg_size[13] + 1 
        if len(cc) == 8:
            deg_dict = {key: value for (key, value) in list(G_full.degree(list(cc)))}
            if (np.array(list(deg_dict.values())) <= degree_cutoff).sum() >= 2:
                seg_size[14] = seg_size[14] + 1
            else:
                seg_size[15] = seg_size[15] + 1 
        if len(cc) == 9:
            deg_dict = {key: value for (key, value) in list(G_full.degree(list(cc)))}
            if (np.array(list(deg_dict.values())) <= degree_cutoff).sum() >= 2:
                seg_size[16] = seg_size[16] + 1
            else:
                seg_size[17] = seg_size[17] + 1 
        
    return seg_size #(np.array(list(deg_dict.values())) > degree_cutoff).sum()


def remove_egdes(edges, newLabels):
    '''
    Mark boundary of the cut.
    
    :param points: edges: shape (n,); newLabels: shape(k,)

    :rtype: marked edges.
    '''
    if newLabels[int(edges[0])] != newLabels[int(edges[1])]:   
        edges[2] = 0
    else:
        edges[2] = 1
    return edges



def gmm_predict(exp,gmm):
    """Predict the labels for the data samples in X using trained model and replace the zreo using the minimum data label.

    Parameters
    ----------
    exp : array-like, shape (n_samples, n_features)
        List of n_features-dimensional data points. Each row
        corresponds to a single data point.

    gmm : Gussian Mixture model

    Returns
    -------
    labels : array, shape (n_samples,)
        Component labels.
    """
    a=exp[exp>0]
    gmm_pred=gmm.predict(a.reshape(-1,1))
    zero_label=gmm.predict(np.min(a).reshape(-1,1))[0]
    label_pred = gmm.predict(exp.reshape(-1,1))
    if len(np.where(exp == 0)[0]) > 0:
        np.place(label_pred, exp==0, zero_label)
    gmm_pred =label_pred
    return gmm_pred



## 20200629 # new 2021112
def compute_spatial_genomewise_optimize(locs, data_norm, cellGraph, gmmDict, w, n, smooth_factor=60, 
                                         unary_scale_factor=100, label_cost=10, algorithm='expansion'):
    
    
    genes = list()
    nodes = list()
    p_values = list()
    smooth_factors = list()
    pred_labels = list()
    model_list = list()
    model_labels = list()
   
    for geneID in data_norm.columns:
#         print(geneID)
        use_otsu = False
        exp = data_norm.loc[:,geneID].values
        gmm=gmmDict[geneID]
        mi=Moran(exp[n],w,permutations=0)
        zero_part_p=calc_zero_part_p(locs,cellGraph,exp)
        zero_inflated=np.sum(exp==0)

        if mi.I < 0.05:
            start_sf=min(smooth_factor*2,100)
        
        elif mi.I < 0.1:
            start_sf=smooth_factor
        else:
            start_sf=10
            ## decide to use otsu or gmm
            if zero_inflated > len(exp)/5:
                newLabels, thresholds,label_pred = cut_graph_general_otsu(cellGraph, exp, unary_scale_factor, 
                                start_sf, label_cost, algorithm)

                p, node, com = compute_p_CSR_otsu(locs, newLabels , label_pred, exp, cellGraph)
                num_isolate = count_isolate(locs,cellGraph, newLabels) 
                noise_size = sum(num_isolate)
                noise_size_norm = noise_size*(200/len(exp))
                logP = (0-np.log10(min(p)))
                obj_otsu= logP-noise_size_norm

                newLabels,_ = cut_graph_general(cellGraph, exp, gmm, unary_scale_factor, 
                                   start_sf, label_cost, algorithm)
                p, node, com = compute_p_CSR(locs, newLabels, gmm, exp, cellGraph)
                num_isolate = count_isolate(locs,cellGraph, newLabels) 
                noise_size = sum(num_isolate)
                noise_size_norm = noise_size*(200/len(exp))
                logP = (0-np.log10(min(p)))
                obj_gmm= logP-noise_size_norm

                if obj_otsu - obj_gmm > 8 :
                    use_otsu = True


        if (zero_part_p < 1e-4 and zero_inflated > len(exp)/3 and len(exp)<1000) or use_otsu:
            model = 'otsu'
            p_best,node_best,newLabels_best,temp_factor_best,com_best, label_pred_best =compute_single_gene_otsu(
                                        locs, exp, cellGraph, smooth_factor=start_sf, 
                                         unary_scale_factor=100, label_cost=10, algorithm='expansion')
        else:
            model = 'gmm'
            p_best,node_best,newLabels_best,temp_factor_best,com_best, label_pred_best =compute_single_gene_gmm(
                                        locs, exp, cellGraph, gmm, smooth_factor=start_sf, 
                                         unary_scale_factor=100, label_cost=10, algorithm='expansion')

            
        if len(p_best)>0:
            final_factor = temp_factor_best
            p_values.append(p_best)
            nodes.append(node_best)
            genes.append(geneID)
            smooth_factors.append(final_factor)
            pred_labels.append(newLabels_best)
            model_list.append(model)
            model_labels.append(list(label_pred_best))
            
    return nodes, p_values, genes, smooth_factors, pred_labels,model_list, model_labels


def identify_spatial_genes(locs, data_norm, cellGraph, gmmDict, smooth_factor=60,
                      unary_scale_factor=100, label_cost=10, algorithm='expansion',
                      ncores = None):

#    pool = mp.Pool()
    '''
    main function to identify spatially variable genes
    :param file:locs: spatial coordinates (n, 2); data_norm: normalized gene expression;
        smooth_factor=10; unary_scale_factor=100; label_cost=10; algorithm='expansion' 
    :rtype: prediction: a dataframe
    '''    
    # 20211119 coco
    G_full = nx.Graph()
    G_cut = nx.Graph()
    tempGraph = cellGraph.copy()
    G_full.add_nodes_from(list(set(list(tempGraph[:,0].astype(np.int32)) + list(tempGraph[:,1].astype(np.int32)))))
    G_full.add_edges_from(tempGraph[:, 0:2].astype(np.int32)) 
    w=pysal.lib.weights.W.from_networkx(G_full) 
    n=list(w.weights.keys())
    
    all_cores = mp.cpu_count()
    if ncores !=None:
        num_cores = ncores
    else:
        num_cores = int(all_cores*0.5)
    if num_cores > math.floor(data_norm.shape[1]/2):
         num_cores=int(math.floor(data_norm.shape[1]/2))
    print(f'scGCO used {num_cores} out of {all_cores} cores')
    ttt = np.array_split(data_norm,num_cores,axis=1)
    tuples = [(l, d, c, g,ww,nn, s, u, b, a) for l, d, c, g,ww,nn, s, u, b, a in zip(
                                    repeat(locs, num_cores), 
                                    ttt,
                                    repeat(cellGraph, num_cores),
                                    repeat(gmmDict, num_cores),
                                    repeat(w, num_cores),
                                    repeat(n, num_cores),
                                    repeat(smooth_factor, num_cores),
                                    repeat(unary_scale_factor, num_cores), 
                                    repeat(label_cost, num_cores),
                                    repeat(algorithm, num_cores))] 
    
    results = parmap.starmap(compute_spatial_genomewise_optimize, tuples,
                                pm_processes=num_cores, pm_pbar=True)
    
#    pool.close()
# p_values, genes, diff_p_values, exp_diff, smooth_factors, pred_labels, model_results
    nnn = [results[i][0] for i in np.arange(len(results))]
    nodes = reduce(operator.add, nnn)
    ppp = [results[i][1] for i in np.arange(len(results))]
    p_values=reduce(operator.add, ppp)
    ggg = [results[i][2] for i in np.arange(len(results))]
    genes = reduce(operator.add, ggg)
    # exp_ppp = [results[i][3] for i in np.arange(len(results))]
    # exp_pvalues = reduce(operator.add, exp_ppp)  
    # exp_ddd = [results[i][4] for i in np.arange(len(results))]
    # exp_diffs = reduce(operator.add, exp_ddd)      
    fff = [results[i][3] for i in np.arange(len(results))]
    s_factors = reduce(operator.add, fff)
    lll = [results[i][4] for i in np.arange(len(results))]
    pred_labels = reduce(operator.add, lll)
    ml = [results[i][5] for i in np.arange(len(results))]
    model_list = reduce(operator.add, ml)
    mmm = [results[i][6] for i in np.arange(len(results))]
    model_labels = reduce(operator.add, mmm)

    best_p_values=[min(i) for i in p_values]
    fdr = multi.multipletests(np.array(best_p_values), method='fdr_bh')[1]
    #exp_fdr = multi.multipletests(np.array(exp_pvalues), method='fdr_bh')[1]    
    
    labels_array = np.array(pred_labels).reshape(len(genes), pred_labels[0].shape[0])
    data_array = np.array((genes, p_values, fdr, s_factors, nodes, model_list , model_labels), dtype=object).T
    t_array = np.hstack((data_array, labels_array))
    c_labels = ['p_value', 'fdr',  'smooth_factor', 'nodes','model','model_labels']
    for i in np.arange(labels_array.shape[1]) + 1:
        temp_label = 'label_cell_' + str(i)
        c_labels.append(temp_label)
    result_df = pd.DataFrame(t_array[:,1:], index=t_array[:,0], 
                      columns=c_labels)
    
    return result_df

def compute_single_gene_gmm(locs, exp, cellGraph, gmm, smooth_factor=10, 
                           unary_scale_factor=100, label_cost=10, algorithm='expansion'):
        
    size_factor=200
    noise_size_estimate= 9
    start_factor=smooth_factor
    ## 0. init cuts graph
    temp_factor = start_factor
    newLabels, label_pred = cut_graph_general(cellGraph, exp, gmm, unary_scale_factor, 
                                        temp_factor, label_cost, algorithm)
    p, node, com = compute_p_CSR(locs, newLabels, gmm, exp, cellGraph)
    num_isolate = count_isolate(locs,cellGraph, newLabels) 

    # noise_size_inside = sum([num_isolate[num] for num in np.arange(1,2*noise_size_estimate, 2)]) # 0629, 16:16
    # noise_size_border = sum([num_isolate[num] for num in np.arange(0,2*noise_size_estimate, 2)])
    # noise_size= noise_size_inside + noise_size_border
    noise_size = sum(num_isolate)

    noise_size_norm = noise_size*(size_factor/len(exp))
    
    logP = (0-np.log10(min(p))) #changed by wanwan

    obj_val = logP- noise_size_norm
            
    # 1. cuts too many noise
    sf_list=[temp_factor]
    obj_val_list=[obj_val]
    noise_size_norm_list=[noise_size_norm]

    while  (temp_factor < 30) or (obj_val < 0 and len(p)>0 and min(p) < 0.01):
        if noise_size_norm >= 20:
            incre = 20
        else:
            incre = 10
        temp_factor = temp_factor + incre   ## can speed up with +10
        newLabels, label_pred = cut_graph_general(cellGraph, exp, gmm, unary_scale_factor, 
                                        temp_factor, label_cost, algorithm)
        p, node, com = compute_p_CSR(locs, newLabels, gmm, exp, cellGraph)  
        num_isolate = count_isolate(locs,cellGraph, newLabels)  

        # noise_size_inside = sum([num_isolate[num] for num in np.arange(1,2*noise_size_estimate, 2)])
        # noise_size_border = sum([num_isolate[num] for num in np.arange(0,2*noise_size_estimate, 2)])
        # noise_size= noise_size_inside + noise_size_border
        noise_size = sum(num_isolate)
        noise_size_norm = noise_size*(size_factor/len(exp))

        logP = (0-np.log10(min(p))) #changed by wanwan

        obj_val = logP- noise_size_norm
        
        sf_list.append(temp_factor)
        obj_val_list.append(obj_val)
        noise_size_norm_list.append(noise_size_norm)
        if temp_factor > 120:
            break
    
    if len(obj_val_list)>=2:
        temp_factor = sf_list[np.argmax(obj_val_list)]
        
        newLabels, label_pred = cut_graph_general(cellGraph, exp, gmm, unary_scale_factor, 
                                        temp_factor, label_cost, algorithm)
        p, node, com = compute_p_CSR(locs, newLabels, gmm, exp, cellGraph) 
        obj_val = obj_val_list[np.argmax(obj_val_list)]
        noise_size_norm = noise_size_norm_list[np.argmax(obj_val_list)]

    
    p_best = p
    node_best = node
    newLabels_best = newLabels
    temp_factor_best = temp_factor  
    obj_val_best = obj_val
    noise_size_best = noise_size_norm
    com_best=com
    label_pred_best = label_pred
    
    
    ## 4. For small significate p_value pattern, need bigger sf
    if len(p_best)>0 and len(node_best[np.argmin(p_best)]) < noise_size_estimate:
        p = p_best
        node = node_best
        newLabels = newLabels_best
        temp_factor = temp_factor_best      
        obj_val = obj_val_best
        noise_size_norm = noise_size_best
        com =com_best
        label_pred = label_pred_best

        while min(p) < 0.01: 
            temp_factor = temp_factor + 50   
            newLabels, label_pred = cut_graph_general(cellGraph, exp, gmm, unary_scale_factor, 
                                        temp_factor, label_cost, algorithm)
            p, node, com = compute_p_CSR(locs, newLabels, gmm, exp, cellGraph)  
            if temp_factor > 500:
                break

        p_best = p
        node_best = node
        newLabels_best = newLabels
        temp_factor_best = temp_factor      
        com_best=com
        label_pred_best = label_pred
    
    return p_best,node_best,newLabels_best,temp_factor_best, com_best, label_pred_best


def compute_single_gene_otsu(locs, exp, cellGraph, smooth_factor=10, 
                            unary_scale_factor=100, label_cost=10, algorithm='expansion'):
    size_factor=200
    noise_size_estimate= 9
    start_factor=smooth_factor
    ## 0. init cuts graph
    temp_factor = start_factor
        
    ## 0. init cuts graph
    newLabels, thresholds, label_pred = cut_graph_general_otsu(cellGraph, exp, unary_scale_factor, 
                                temp_factor, label_cost, algorithm)

    p, node, com = compute_p_CSR_otsu(locs, newLabels , label_pred, exp, cellGraph)
    num_isolate = count_isolate(locs,cellGraph, newLabels) 

    # noise_size_inside = sum([num_isolate[num] for num in np.arange(1,2*noise_size_estimate, 2)])
    # noise_size_border = sum([num_isolate[num] for num in np.arange(0,2*noise_size_estimate, 2)])
    # noise_size= noise_size_inside + noise_size_border
    noise_size = sum(num_isolate)
    noise_size_norm = noise_size*(size_factor/len(exp))  
    
    logP = (0-np.log10(min(p))) #changed by wanwan

    obj_val = logP- noise_size_norm

    sf_list=[temp_factor]
    obj_val_list=[obj_val]
    noise_size_norm_list=[noise_size_norm]

    while  (temp_factor < 30) or (obj_val < 1 and len(p)>0 and min(p) < 0.01):
        if noise_size_norm >= 20:
            incre = 20
        else:
            incre = 10
        temp_factor = temp_factor + incre   ## can speed up with +10
        newLabels, thresholds,label_pred = cut_graph_general_otsu(cellGraph, exp, unary_scale_factor, 
                                temp_factor, label_cost, algorithm)

        p, node, com = compute_p_CSR_otsu(locs, newLabels , label_pred, exp, cellGraph)
        num_isolate = count_isolate(locs,cellGraph, newLabels)  

        # noise_size_inside = sum([num_isolate[num] for num in np.arange(1,2*noise_size_estimate, 2)])
        # noise_size_border = sum([num_isolate[num] for num in np.arange(0,2*noise_size_estimate, 2)])
        # noise_size= noise_size_inside + noise_size_border
        noise_size = sum(num_isolate)
        noise_size_norm = noise_size*(size_factor/len(exp))

        logP = (0-np.log10(min(p))) #changed by wanwan

        obj_val = logP- noise_size_norm
        
        sf_list.append(temp_factor)
        obj_val_list.append(obj_val)
        noise_size_norm_list.append(noise_size_norm)
        if temp_factor > 120:
            break
    
    if len(obj_val_list)>=2:
        temp_factor = sf_list[np.argmax(obj_val_list)]
        
        newLabels, thresholds,label_pred = cut_graph_general_otsu(cellGraph, exp, unary_scale_factor, 
                                temp_factor, label_cost, algorithm)

        p, node, com = compute_p_CSR_otsu(locs, newLabels , label_pred, exp, cellGraph)
        obj_val = obj_val_list[np.argmax(obj_val_list)]
        noise_size_norm = noise_size_norm_list[np.argmax(obj_val_list)]

    p_best = p
    node_best = node
    newLabels_best = newLabels
    temp_factor_best = temp_factor  
    obj_val_best = obj_val
    noise_size_best = noise_size_norm
    com_best=com
    label_pred_best = label_pred
    
    ## 4. For small significate p_value pattern, need bigger sf
    if len(p_best)>0 and len(node_best[np.argmin(p_best)]) < noise_size_estimate:
        p = p_best
        node = node_best
        newLabels = newLabels_best
        temp_factor = temp_factor_best      
        obj_val = obj_val_best
        noise_size_norm = noise_size_best
        com =com_best
        label_pred = label_pred_best

        while min(p) < 0.01:
            temp_factor = temp_factor + 50   
            newLabels, thresholds, label_pred = cut_graph_general_otsu(cellGraph, exp, unary_scale_factor, 
                                temp_factor, label_cost, algorithm)

            p, node, com = compute_p_CSR_otsu(locs, newLabels , label_pred, exp, cellGraph)
            if temp_factor > 500:
                break
            
        p_best = p
        node_best = node
        newLabels_best = newLabels
        temp_factor_best = temp_factor      
        com_best=com
        label_pred_best = label_pred
    
    return p_best,node_best,newLabels_best,temp_factor_best,com_best, label_pred_best


def identify_spatial_genes_fixed_sf(locs, data_norm, cellGraph, gmmDict, smooth_factor=30,
                      unary_scale_factor=100, label_cost=10, algorithm='expansion',
                      model = 'gmm'):

#    pool = mp.Pool()
    '''
    main function to identify spatially variable genes
    :param file:locs: spatial coordinates (n, 2); data_norm: normalized gene expression;
        smooth_factor=10; unary_scale_factor=100; label_cost=10; algorithm='expansion' 
    :rtype: prediction: a dataframe
    '''    
    
    num_cores = int(mp.cpu_count()/2)
    if num_cores > math.floor(data_norm.shape[1]/2):
         num_cores=int(math.floor(data_norm.shape[1]/2))
    ttt = np.array_split(data_norm,num_cores,axis=1)
    tuples = [(l, d, c, g, s, u, b, a, m) for l, d, c, g, s, u, b, a, m in zip(
                                    repeat(locs, num_cores), 
                                    ttt,
                                    repeat(cellGraph, num_cores),
                                    repeat(gmmDict, num_cores),
                                    repeat(smooth_factor, num_cores),
                                    repeat(unary_scale_factor, num_cores), 
                                    repeat(label_cost, num_cores),
                                    repeat(algorithm, num_cores),
                                    repeat(model, num_cores))] 
    
    results = parmap.starmap(compute_single_fixed_sf, tuples,
                                pm_processes=num_cores, pm_pbar=True)
    
#    pool.close()
# p_values, genes, diff_p_values, exp_diff, smooth_factors, pred_labels, model_results
    nnn = [results[i][0] for i in np.arange(len(results))]
    nodes = reduce(operator.add, nnn)
    ppp = [results[i][1] for i in np.arange(len(results))]
    p_values=reduce(operator.add, ppp)
    ggg = [results[i][2] for i in np.arange(len(results))]
    genes = reduce(operator.add, ggg)
    # exp_ppp = [results[i][3] for i in np.arange(len(results))]
    # exp_pvalues = reduce(operator.add, exp_ppp)  
    # exp_ddd = [results[i][4] for i in np.arange(len(results))]
    # exp_diffs = reduce(operator.add, exp_ddd)      
    fff = [results[i][3] for i in np.arange(len(results))]
    s_factors = reduce(operator.add, fff)
    lll = [results[i][4] for i in np.arange(len(results))]
    pred_labels = reduce(operator.add, lll)
    ml = [results[i][5] for i in np.arange(len(results))]
    model_list = reduce(operator.add, ml)
    mmm = [results[i][6] for i in np.arange(len(results))]
    model_labels = reduce(operator.add, mmm)

    best_p_values=[min(i) for i in p_values]
    fdr = multi.multipletests(np.array(best_p_values), method='fdr_bh')[1]
    
    labels_array = np.array(pred_labels).reshape(len(genes), pred_labels[0].shape[0])
    data_array = np.array((genes, p_values, fdr, s_factors, nodes, model_list , model_labels), dtype=object).T
    t_array = np.hstack((data_array, labels_array))
    c_labels = ['p_value', 'fdr',  'smooth_factor', 'nodes','model','model_labels']
    for i in np.arange(labels_array.shape[1]) + 1:
        temp_label = 'label_cell_' + str(i)
        c_labels.append(temp_label)
    result_df = pd.DataFrame(t_array[:,1:], index=t_array[:,0], 
                      columns=c_labels)
    
    return result_df

def compute_single_fixed_sf(locs, data_norm, cellGraph,gmmDict, smooth_factor=30, 
                        unary_scale_factor=100,label_cost = 10, algorithm ='expansion', 
                        model =  'gmm',**kw):
    '''
    default model is GMM, parameters
    '''
    
    genes = list()
    nodes = list()
    p_values = list()
    smooth_factors = list()
    pred_labels = list()
    model_list = list()
    model_labels = list()
    smooth_factor = smooth_factor
    for geneID in  data_norm.columns:
        exp = data_norm.loc[:,geneID].values
       
        if model == 'otsu':
            p_best,node_best,newLabels_best,temp_factor_best,com_best, label_pred_best = compute_single_otsu_fixed_sf(
                                        locs,exp,cellGraph,smooth_factor, 
                                        unary_scale_factor, label_cost, algorithm)
        else:
            model = 'gmm' 
            gmm = gmmDict[geneID]
          
            p_best,node_best,newLabels_best,temp_factor_best,com_best, label_pred_best = compute_single_gmm_fixed_sf(
                    locs,exp,cellGraph,gmm, smooth_factor, 
                    unary_scale_factor, label_cost, algorithm)

        final_factor = smooth_factor
        p_values.append(p_best)
        nodes.append(node_best)
        genes.append(geneID)
        smooth_factors.append(final_factor)
        pred_labels.append(newLabels_best)
        model_list.append(model)
        model_labels.append(list(label_pred_best))
            
    return nodes, p_values, genes, smooth_factors, pred_labels, model_list, model_labels


def compute_single_gmm_fixed_sf(locs,exp,cellGraph,gmm, smooth_factor=30, 
                    unary_scale_factor=100, label_cost=10, algorithm='expansion'):
    
    newLabels, label_pred = cut_graph_general(cellGraph, exp, gmm, unary_scale_factor, 
                                        smooth_factor, label_cost, algorithm)
    p, node, com = compute_p_CSR(locs, newLabels, gmm, exp, cellGraph) 
    
    p_best = p
    node_best = node
    newLabels_best = newLabels
    temp_factor_best = smooth_factor
    com_best=com
    label_pred_best = label_pred

    return p_best,node_best,newLabels_best,temp_factor_best,com_best, label_pred_best


def compute_single_otsu_fixed_sf(locs,exp,cellGraph,smooth_factor=30, 
                          unary_scale_factor=100, label_cost=10, algorithm='expansion'):
    
    newLabels, thresholds,label_pred = cut_graph_general_otsu(cellGraph, exp, unary_scale_factor, 
                                smooth_factor, label_cost, algorithm)

    p, node, com = compute_p_CSR_otsu(locs, newLabels , label_pred, exp, cellGraph)
    
    p_best = p
    node_best = node
    newLabels_best = newLabels
    temp_factor_best = smooth_factor
    com_best=com
    label_pred_best = label_pred

    return p_best,node_best,newLabels_best,temp_factor_best,com_best, label_pred_best


# implement this one
def compute_p_CSR_otsu(locs, newLabels , label_pred, exp, cellGraph): 
    '''
    Returns p_value of the cut with otsu.
    
    :param points: newLabels: shape (n,); gmm: gmm object
                   exp: ndarray shape (n ,3); cellGraph: shape (n,3)

    :rtype: p_value.
    '''
    
    p_values = list()
    node_lists = list()
    label_pred = label_pred
    noise_size_estimate=9 
    unique, counts = np.unique(label_pred,return_counts=True)
    con_components = count_component(locs,cellGraph, newLabels) 
    noise = dict()
    # now calculate p for all comp without considering noise
    min_sig_p_size = np.inf
    for j in np.arange(len(con_components)):
        node_list = con_components[j]
        com_size = len(node_list)
        if len(exp)<= 1000:
            com_cutoff = 3
            addNoise = True
        else:
            com_cutoff = 9
            addNoise = False
        if com_size >= com_cutoff:
            temp_exp = exp[np.array(list(node_list))]
            label_pred_com = label_pred[np.array(list(node_list))]
        # check 0s
            unique_com, counts_com = np.unique(label_pred_com, return_counts=True)  
            major_label = unique_com[np.where(counts_com == counts_com.max())[0][0]]
            label_count = counts[np.where(unique == major_label)[0]]  ##  real counts that in subgraphs,Ci
            count_in_com =  counts_com.max()   ## counts by graph cuts,k
            cover = exp.shape[0]/com_size
            prob_cdf = (1 - poisson.cdf(count_in_com, com_size*(label_count/exp.shape[0]))[0]) #*cover
            
            psf = poisson.sf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
            pmf = poisson.pmf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
            prob=min((psf+pmf),1)  
            if prob < 10e-30:
                prob = 10e-30
            p_values.append(prob)
            if prob <0.1 and len(con_components[j]) < min_sig_p_size:
                min_sig_p_size = len(con_components[j])
        else: # set small comp p=1
            p_values.append(1)            
        node_lists.append(np.array(list(node_list)))
#        print(p_values)
    if addNoise:
        for j in np.arange(len(con_components)):
            if p_values[j] >= 0.1 and len(con_components[j]) < 10: # min_sig_p_size:
                noise[j] = con_components[j]
        # now re-calculate p considering noise

        for j in np.arange(len(con_components)):
            if p_values[j] < 0.1: # small p let correct by consider noise
                noise_size = 0
                used_com = list()
                for jj, cc in noise.items():
                    if noise_inside(cc, con_components[j], cellGraph):
                        noise_size = noise_size + len(cc)
                        used_com.append(jj)
                if noise_size > 0:
                    for jj in used_com:
                        noise.pop(jj)
                    node_list = con_components[j]
                    com_size = len(node_list)
                    temp_exp = exp[np.array(list(node_list))]
                    label_pred_com = label_pred[np.array(list(node_list))]
                    unique_com, counts_com = np.unique(label_pred_com, return_counts=True)  
                    major_label = unique_com[np.where(counts_com == counts_com.max())[0][0]]
                    label_count = counts[np.where(unique == major_label)[0]]  ##  real counts that in subgraphs,Ci
                    count_in_com =  counts_com.max()   ## counts by graph cuts,k
                    com_size = com_size + noise_size
                    cover = exp.shape[0]/com_size
                    prob_cdf = (1 - poisson.cdf(count_in_com, com_size*(label_count/exp.shape[0]))[0]) #*cover
                
                    psf = poisson.sf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
                    pmf = poisson.pmf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
                    prob = min((psf + pmf),1)   
                    if prob < 10e-30:
                        prob = 10e-30
                    p_values[j] = prob      
                    
    return p_values, node_lists, con_components

## try modify 20200602 
def compute_unary_cost_simple_otsu(count, thresholds, unary_scale_factor = 100):
    '''
    Returns unary cost energy.
    
    :param points: count: shape (n,); gmm: gmm object; scale_factor: scalar

    :rtype: unary energy matrix.
    '''    
   
    labels_pred = np.zeros(count.shape[0])
    if isinstance(thresholds, float):
        labels_pred[np.where(count > thresholds)] = 1
        mid_points = np.zeros(1)
        temp_means = np.zeros(2)
        temp_covs = np.zeros(2)
    else:
        for i in np.arange(thresholds.shape[0]):
            labels_pred[np.where(count > thresholds[i])] = i + 1      
            mid_points = np.zeros(thresholds.shape[0])
            temp_means = np.zeros(thresholds.shape[0] + 1)
            temp_covs = np.zeros(thresholds.shape[0]+1)

    for i in np.arange(len(temp_means)):
        temp_means[i] = np.mean(count[labels_pred == i])
        temp_covs[i] = np.var(count[labels_pred==i], ddof=1)

    for i in np.arange(len(mid_points)):
        #  mid_points[i] = np.max(count[labels_pred == i])
         mid_points[i] = (temp_means[i]*np.sqrt(temp_covs[i+1]) + 
                     temp_means[i+1]*np.sqrt(temp_covs[i])
                    )/(np.sqrt(temp_covs[i]) + np.sqrt(temp_covs[i+1]))
    temp = count[:, np.newaxis] - temp_means.T[1:]
    neg_indices = np.apply_along_axis(first_neg_index, 1, temp)
    ind_count_arr = np.vstack((neg_indices, count)).T        
    unary_cost =  (unary_scale_factor*np.apply_along_axis(calc_u_cost, 1, 
                                    ind_count_arr, mid_points)).astype(np.int32)  

    a = count.copy()
    if sum(a>1)/len(count) <= 0.1 and sum(a>1)<=30:
        unique_cost, counts_cost = np.unique(unary_cost, return_counts=True)
        val_to_replace1 = unique_cost[np.argmax(counts_cost)]
        val_to_replace2 = 0 - unique_cost[np.argmax(counts_cost)]
        if val_to_replace1 < 10 and val_to_replace1 > 0:
            np.place(unary_cost, unary_cost == val_to_replace1, np.median(abs(unique_cost)))
        elif val_to_replace1 < 0 and val_to_replace1 > -10:
            np.place(unary_cost, unary_cost == val_to_replace1, 0-np.median(abs(unique_cost)))
        if val_to_replace2 < 10 and val_to_replace2 > 0:
            np.place(unary_cost, unary_cost == val_to_replace2, np.median(abs(unique_cost)))
        elif val_to_replace2 < 0 and val_to_replace2 > -10:
            np.place(unary_cost, unary_cost == val_to_replace2, 0-np.median(abs(unique_cost)))       
    return labels_pred.astype(np.int32), unary_cost

## wang
def compute_unary_cost_simple_otsu_mid(count, thresholds, unary_scale_factor = 100):
    '''
    depprecated

    Returns unary cost energy.
    
    :param points: count: shape (n,); gmm: gmm object; scale_factor: scalar

    :rtype: unary energy matrix.
    '''    
   
    labels_pred = np.zeros(count.shape[0])
    if isinstance(thresholds, float):
        labels_pred[np.where(count > thresholds)] = 1
        mid_points = np.zeros(1)
        temp_means = np.zeros(2)
    else:
        for i in np.arange(thresholds.shape[0]):
            labels_pred[np.where(count > thresholds[i])] = i + 1      
    
            mid_points = np.zeros(thresholds.shape[0])
            temp_means = np.zeros(thresholds.shape[0] + 1)
   

    for i in np.arange(len(temp_means)):
        temp_means[i] = np.mean(count[labels_pred == i])
 
    for i in np.arange(len(mid_points)):
         mid_points[i] = np.max(count[labels_pred == i])
    
    temp = count[:, np.newaxis] - mid_points.T 
    neg_indices = np.apply_along_axis(first_neg_index, 1, temp)
    ind_count_arr = np.vstack((neg_indices, count)).T               
    return labels_pred.astype(np.int32), (unary_scale_factor*np.apply_along_axis(calc_u_cost, 1, 
                                    ind_count_arr, mid_points)).astype(np.int32)


def compute_unary_cost_simple_profile_otsu(count, thresholds, unary_scale_factor):
    '''
    Returns unary cost energy just use profile without exp.
    
    :param points: count: shape (n,); gmm: gmm object; scale_factor: scalar

    :rtype: unary energy matrix.
    '''    
    labels_pred = np.zeros(count.shape[0])
    if isinstance(thresholds, float):
        labels_pred[np.where(count > thresholds)] = 1       
    else:
        for i in np.arange(thresholds.shape[0]):
            labels_pred[np.where(count > thresholds[i])] = i + 1

    uniq, count = np.unique(labels_pred, return_counts = True)    

    unary_mat = np.zeros((len(labels_pred), len(uniq)))
    for i in np.arange(unary_mat.shape[0]):
        for j in np.arange(len(uniq)):
            if uniq[j] == labels_pred[i]:  ## same ,energy -1; imsame ,energy 1.
                unary_mat[i, j] = -1
            else:
                unary_mat[i, j] = 1   
    return labels_pred.astype(np.int32), (unary_scale_factor*unary_mat).astype(np.int32)


def cut_graph_general_otsu(cellGraph, exp, unary_scale_factor=100, 
                      smooth_factor=10, label_cost=10, algorithm='expansion',
                      profile=False):
    '''
    Returns new labels and gmm for the cut.
    
    :param points: cellGraph (n,3); count: shape (n,); 
    :unary_scale_factor, scalar; smooth_factor, scalar; 
    :label_cost: scalar; algorithm='expansion'
    :rtype: label shape (n,); gmm object.
    '''
    count=exp
#    thresholds = threshold_multiotsu(exp) #3D
    thresholds = threshold_otsu(count)  #2D
    if isinstance(thresholds, float):
        pair_size = 2
    else:
        pair_size = thresholds.shape[0] + 1
    smooth_factor = smooth_factor
    unary_scale_factor = unary_scale_factor
    label_cost = label_cost
    algorithm = algorithm
    if profile==False:
        label_pred, unary_cost = compute_unary_cost_simple_otsu(count, thresholds, unary_scale_factor)
    else:
        label_pred,unary_cost= compute_unary_cost_simple_profile_otsu(count, thresholds, unary_scale_factor)
    pairwise_cost = compute_pairwise_cost(pair_size, smooth_factor)
    edges = cellGraph[:,0:2].astype(np.int32)
    labels = pygco.cut_from_graph(edges, unary_cost, pairwise_cost, label_cost)
#    energy = compute_energy(unary_cost, pairwise_cost, edges, labels)

    return labels, thresholds ,label_pred


def calc_zero_part_p(locs, cellGraph, exp):
    temp_exp = exp*1000
    temp_newLabels = temp_exp.astype(int)

    con_components = count_component(locs,cellGraph, temp_newLabels) 

    noise = dict()
    # now calculate p for all comp without considering noise
    if True:
        for j in np.arange(len(con_components)):
            if j > 0:
                break
            node_list = con_components[j]
            com_size = len(node_list)
            label_count = len(np.where(temp_newLabels == 0)[0])
            count_in_com = com_size
            prob_cdf = 1 - poisson.cdf(count_in_com, com_size*(label_count/exp.shape[0]))
            cover = exp.shape[0]/com_size

            if label_count==0:
                prob_sf=1
            else:
                psf = poisson.sf(count_in_com, com_size*(label_count/exp.shape[0]))
                pmf = poisson.pmf(count_in_com, com_size*(label_count/exp.shape[0]))
                prob_sf = (psf + pmf)
            #print(count_in_com, com_size, label_count, exp.shape[0])
    return min(prob_sf*cover,1)




import multiprocessing as mp
from sklearn.cluster import DBSCAN,KMeans
import numpy as np
import pandas as pd
from scipy.spatial.distance import cdist, hamming
import math
import parmap
import random
import parmap
import itertools
from itertools import repeat
from scipy.spatial import distance
import operator
from tqdm import tqdm
from functools import reduce
from sklearn import mixture
import statsmodels.stats.multitest as multi
import networkx as nx
from .Preprocessing import *
from .Graph_cut import *
from .Visualization import *
    
    
def calc_distance_com_single(geneID,nodes,model_labels, locs, data_norm,tissue_mat):
    genelist = []
    recal_genelist = []
    ham_dist_list = []
    ham_new_dist_list = []
    jac_dist_list = []
    haus_dist_list = []
    
    if True:
        if geneID in data_norm.columns:
            node = nodes
            model_label = model_labels
            if len(node)==1 and len(node[0])==locs.shape[0]:
                recal_genelist.append(geneID)
            else:
                # print(geneID)
                genelist.append(geneID)
                target_mat = list()
                target_class = []
                for nn in range(len(node)):
                    pred_com = model_label[np.array(node[nn])]  
                    unique_com, counts_com = np.unique(pred_com, return_counts=True)
                    major_label = unique_com[np.where(counts_com == counts_com.max())[0][0]]
                    temp_target = np.zeros(locs.shape[0])
                    temp_target[node[nn]] = 1
                    if sum(temp_target)< (1/2)*locs.shape[0]+10:
                        # print(nn, len(node[nn]))
                        target_mat.append(temp_target)
                        target_class.append(major_label)

                overlap = cdist(tissue_mat, target_mat,compute_inclusion_min)
                
                # merge these best matched target whose are one group gmmLabels.
                if len(target_mat)>1:  # target mat matching one tissue mat were merged
                    if sum(np.max(overlap, axis = 0)>0)>0:
                        maxInd = np.argmax(overlap, axis=0)
                        Ind, num = np.unique(np.argmax(overlap.T[np.max(overlap, axis= 0)>0].T, axis=0),
                                            return_counts=True)
                        if np.any(np.max(overlap, axis = 0) == 0):
                            np.place(maxInd, np.max(overlap, axis = 0) == 0, -1)  ## just Exclude those that don't match at all
                        total_targe_mat = []
                        if np.any(np.array(num)==1):
                            inds = np.array(Ind)[np.array(num)==1]
                            for ind in inds:
                                temp_targe_mat = np.array(target_mat)[maxInd == ind]
                                total_targe_mat.append(temp_targe_mat[0])
                                # print('Rank{} No need merge targe: {}'.format(ind, np.where(maxInd == ind)[0]))

                        if np.any(np.array(num)>1):
                            inds = np.array(Ind)[np.array(num)>1]
                            for ind in inds:
                                for cl in np.unique(target_class):
                                    merge_ind = set(np.where(np.array(target_class) == cl)[0])& set(np.where(maxInd == ind)[0])
                                    # print('Rank{} & Labels{} Need merge targe:{}'.format(ind, cl, merge_ind))
                                    if len(merge_ind)>0:
                                        temp_targe_mat = sum(np.array(target_mat)[np.array(list(merge_ind))])
                                        if sum(temp_targe_mat>1)>0:
                                            np.place(temp_targe_mat, temp_targe_mat>1, 1)
                                        if sum(temp_targe_mat) > locs.shape[0]/2+10:
                                            temp_targe_mat = abs(1 - temp_targe_mat)
                                        total_targe_mat.append(temp_targe_mat)
                #             print(inds, len(total_targe_mat))
                    else:
                        total_targe_mat = target_mat
                else:
                    total_targe_mat = target_mat

                # merge tissue 
                if len(tissue_mat)>1:  # tissue mat matching one targe mat were merged
                    maxInd = np.argmax(overlap, axis=1)
                    if sum(np.max(overlap, axis = 1)>0)>0:
                        Ind_tis, num_tis = np.unique(np.argmax(overlap[np.max(overlap, axis = 1)>0],axis=1),
                                                    return_counts=True)
                        if np.any(np.max(overlap, axis = 1) == 0):
                            np.place(maxInd, np.max(overlap, axis = 1) == 0, -1)  ## just Exclude those that don't match at all
                        total_tissue_mat = []
                        if np.any(np.array(num_tis)==1):
                            inds = np.array(Ind_tis)[np.array(num_tis)==1]
                            for ind in inds:
                                temp_tissue_mat = np.array(tissue_mat)[maxInd == ind]
                                total_tissue_mat.append(temp_tissue_mat[0])
                                # print('Rank{} No need merge tissue: {}'.format(ind, np.where(maxInd==ind)[0]))

                        if np.any(np.array(num_tis)>1):
                            inds = np.array(Ind_tis)[np.array(num_tis)>1]
                            for ind in inds:
                                temp_tissue_mat = sum(np.array(tissue_mat)[maxInd == ind])
                                if sum(temp_tissue_mat>1)>0:
                                    np.place(temp_tissue_mat, temp_tissue_mat>1, 1)
                                total_tissue_mat.append(temp_tissue_mat)
                                # print('Rank{} Need merge tissue: {}'.format(ind, np.where(maxInd==ind)[0]))
                    else:
                        total_targe_mat = tissue_mat
                else:
                    total_tissue_mat = tissue_mat


                ham_dist = cdist(total_tissue_mat, total_targe_mat, compute_diff_vs_common_const10)
                ham_dist_list.append(np.min(ham_dist))

                ham_new_dist = cdist(total_tissue_mat, total_targe_mat, compute_diff_vs_common)
                ham_new_dist_list.append(np.min(ham_new_dist))

                jac_dist = cdist(total_tissue_mat, total_targe_mat, jaccard_dist)
                jac_dist_list.append(np.min(jac_dist))

                haus_dist = []
                for tar in total_targe_mat:
                    target_locs = locs[tar==1]
                    for tis in total_tissue_mat:
                        tissue_locs = locs[tis==1]
                        haus_dist.append(compute_hausdorff(tissue_locs, target_locs))
                haus_dist_list.append(np.min(haus_dist))

    return genelist, ham_dist_list, jac_dist_list,haus_dist_list, ham_new_dist_list

def calc_distance_com(geneList,nodes,model_labels, locs, data_norm,tissue_mat):
    genelist = []
    recal_genelist = []
    ham_dist_list = []
    ham_new_dist_list = []
    jac_dist_list = []
    haus_dist_list = []
    
    for geneID in geneList:
        if geneID in data_norm.columns:
            node = nodes[geneID]
            model_label = np.array(model_labels[geneID])
            if len(node)==1 and len(node[0])==locs.shape[0]:
                recal_genelist.append(geneID)
            else:
                # print(geneID)
                genelist.append(geneID)
                target_mat = list()
                target_class = []
                for nn in range(len(node)):
                    pred_com = model_label[np.array(node[nn])]  
                    unique_com, counts_com = np.unique(pred_com, return_counts=True)
                    major_label = unique_com[np.where(counts_com == counts_com.max())[0][0]]
                    temp_target = np.zeros(locs.shape[0])
                    temp_target[node[nn]] = 1
                    if sum(temp_target)< (1/2)*locs.shape[0]+10:
                        # print(nn, len(node[nn]))
                        target_mat.append(temp_target)
                        target_class.append(major_label)

                overlap = cdist(tissue_mat, target_mat,compute_inclusion_min)
                
                # merge these best matched target whose are one group gmmLabels.
                if len(target_mat)>1:  # target mat matching one tissue mat were merged
                    if sum(np.max(overlap, axis = 0)>0)>0:
                        maxInd = np.argmax(overlap, axis=0)
                        Ind, num = np.unique(np.argmax(overlap.T[np.max(overlap, axis= 0)>0].T, axis=0),
                                            return_counts=True)
                        if np.any(np.max(overlap, axis = 0) == 0):
                            np.place(maxInd, np.max(overlap, axis = 0) == 0, -1)  ## just Exclude those that don't match at all
                        total_targe_mat = []
                        if np.any(np.array(num)==1):
                            inds = np.array(Ind)[np.array(num)==1]
                            for ind in inds:
                                temp_targe_mat = np.array(target_mat)[maxInd == ind]
                                total_targe_mat.append(temp_targe_mat[0])
                                # print('Rank{} No need merge targe: {}'.format(ind, np.where(maxInd == ind)[0]))

                        if np.any(np.array(num)>1):
                            inds = np.array(Ind)[np.array(num)>1]
                            for ind in inds:
                                for cl in np.unique(target_class):
                                    merge_ind = set(np.where(np.array(target_class) == cl)[0])& set(np.where(maxInd == ind)[0])
                                    # print('Rank{} & Labels{} Need merge targe:{}'.format(ind, cl, merge_ind))
                                    if len(merge_ind)>0:
                                        temp_targe_mat = sum(np.array(target_mat)[np.array(list(merge_ind))])
                                        if sum(temp_targe_mat>1)>0:
                                            np.place(temp_targe_mat, temp_targe_mat>1, 1)
                                        if sum(temp_targe_mat) > locs.shape[0]/2+10:
                                            temp_targe_mat = abs(1 - temp_targe_mat)
                                        total_targe_mat.append(temp_targe_mat)
                #             print(inds, len(total_targe_mat))
                    else:
                        total_targe_mat = target_mat
                else:
                    total_targe_mat = target_mat

                # merge tissue 
                if len(tissue_mat)>1:  # tissue mat matching one targe mat were merged
                    maxInd = np.argmax(overlap, axis=1)
                    if sum(np.max(overlap, axis = 1)>0)>0:
                        Ind_tis, num_tis = np.unique(np.argmax(overlap[np.max(overlap, axis = 1)>0],axis=1),
                                                    return_counts=True)
                        if np.any(np.max(overlap, axis = 1) == 0):
                            np.place(maxInd, np.max(overlap, axis = 1) == 0, -1)  ## just Exclude those that don't match at all
                        total_tissue_mat = []
                        if np.any(np.array(num_tis)==1):
                            inds = np.array(Ind_tis)[np.array(num_tis)==1]
                            for ind in inds:
                                temp_tissue_mat = np.array(tissue_mat)[maxInd == ind]
                                total_tissue_mat.append(temp_tissue_mat[0])
                                # print('Rank{} No need merge tissue: {}'.format(ind, np.where(maxInd==ind)[0]))

                        if np.any(np.array(num_tis)>1):
                            inds = np.array(Ind_tis)[np.array(num_tis)>1]
                            for ind in inds:
                                temp_tissue_mat = sum(np.array(tissue_mat)[maxInd == ind])
                                if sum(temp_tissue_mat>1)>0:
                                    np.place(temp_tissue_mat, temp_tissue_mat>1, 1)
                                total_tissue_mat.append(temp_tissue_mat)
                                # print('Rank{} Need merge tissue: {}'.format(ind, np.where(maxInd==ind)[0]))
                    else:
                        total_targe_mat = tissue_mat
                else:
                    total_tissue_mat = tissue_mat


                ham_dist = cdist(total_tissue_mat, total_targe_mat, compute_diff_vs_common_const10)
                ham_dist_list.append(np.min(ham_dist))

                ham_new_dist = cdist(total_tissue_mat, total_targe_mat, compute_diff_vs_common)
                ham_new_dist_list.append(np.min(ham_new_dist))

                jac_dist = cdist(total_tissue_mat, total_targe_mat, jaccard_dist)
                jac_dist_list.append(np.min(jac_dist))

                haus_dist = []
                for tar in total_targe_mat:
                    target_locs = locs[tar==1]
                    for tis in total_tissue_mat:
                        tissue_locs = locs[tis==1]
                        haus_dist.append(compute_hausdorff(tissue_locs, target_locs))
                haus_dist_list.append(np.min(haus_dist))

    return genelist, ham_dist_list, jac_dist_list,haus_dist_list, ham_new_dist_list

        

def calc_distance_com_parallel(geneList, nodes, model_labels, locs, data_norm, tissue_mat):
    num_cores = mp.cpu_count()
    if num_cores > math.floor(len(geneList)/2):
         num_cores=int(math.floor(len(geneList)/2))
    ttt = np.array_split(np.array(list(geneList)),num_cores)
    tuples = [(g, n, m, l, d, t) for g, n, m, l, d, t in zip( 
                                    ttt,
                                    repeat(nodes, num_cores),
                                    repeat(model_labels, num_cores),
                                    repeat(locs, num_cores),
                                    repeat(data_norm, num_cores),
                                    repeat(tissue_mat, num_cores))] 
    
    results = parmap.starmap(calc_distance_com, tuples,
                                pm_processes=num_cores, pm_pbar=True)
    
    g = [results[i][0] for i in np.arange(len(results))]
    genes = reduce(operator.add, g)
    ham = [results[i][1] for i in np.arange(len(results))]
    ham_dist_list = reduce(operator.add, ham)
    jac = [results[i][2] for i in np.arange(len(results))]
    jac_dist_list = reduce(operator.add, jac)
    haus = [results[i][3] for i in np.arange(len(results))]
    haus_dist_list = reduce(operator.add, haus)
    ham2 = [results[i][4] for i in np.arange(len(results))]
    ham_new_dist_list = reduce(operator.add, ham2)
    
    
    df = pd.DataFrame(index=genes, 
        data= {'Hamming':ham_dist_list, 
                'Jaccard':jac_dist_list,
                'Hausdorff':haus_dist_list,
                'Hamming2':ham_new_dist_list})
    return df


def compute_inclusion_min(u, v):
#    if len(np.where((u+v) == 2)[0]) == 0:
#        return len(np.where(abs(u-v) == 1)[0])
#    else:
    return len(np.where((u+v) == 2)[0])/min(sum(v),sum(u))


def compute_hausdorff(u,v):
    '''
    Compute norm hausdorff geometry distance between two pattern.
    u: target_mat
    v: tissue_mat
    
    '''
    
    dist_u_v=distance.directed_hausdorff(u,v)[0]
    dist_v_u=distance.directed_hausdorff(v,u)[0]
    
    # if norm_factor:
    #     dist= max(dist_u_v/compute_norm_factor(v),dist_v_u/compute_norm_factor(u))
    # else:     
    return max(dist_u_v,dist_v_u)


def compute_diff_vs_common_const10(u, v):
    return len(np.where((u+v) == 1)[0])/(2*len(np.where((u+v) == 2)[0]) + 10)

def compute_diff_vs_common(u, v):
    return len(np.where((u+v) == 1)[0])/(2*len(np.where((u+v) == 2)[0]) + 1)

def jaccard_dist(u, v):
    return  1 - len(np.where((u+v) == 2)[0])/(len(np.where((u+v) > 0)[0]))


def recalc_distance_com(geneList, models, locs, data_norm,cellGraph, tissue_mat, gmmDict):
    
    
    noise_size_estimate =9 # min(np.min(np.sum(cand_seed_mat, axis=1)), 9)  ## #noise<=9
    size_factor=200
    new_hamming=[]
    new_jaccard=[]
    new_hausdorff=[]  
    new_hamming2= []
    
    genes = []
    p_values = list()
    node_list = list()
    s_factors = list()
    model_list = list()
    model_labels = list()
    pred_labels = list()
            
    for geneID in geneList: #zeor_boundGenes: #de_counts: 
        # print(geneID)
        hamming_dist_list=[]
        jaccard_dist_list=[]
        hausdorff_dist_list=[]
        hamming_new_dist_list = []
        p_min=[]
        smooth_factors = list()
       
        exp = data_norm.loc[:,geneID].values
        if geneID in models.index:
            model = models[geneID]
            gmm = gmmDict[geneID]
        else:
            model = 'gmm'
            gmm = perform_gmm(exp)
          
        temp_factor = 0   # 'Cldn9' need start from sf=0 
        if model == 'gmm':
            p,nodes,newLabels,_,com, label_pred = compute_single_gmm_fixed_sf(
                            locs,exp,cellGraph, gmm, temp_factor)
        else:
            p,nodes,newLabels,_,com, label_pred = compute_single_otsu_fixed_sf(
                                        locs,exp,cellGraph,temp_factor)
        num_isolate = count_isolate(locs,cellGraph, newLabels) 

        noise_size = sum(num_isolate[0:int(2*noise_size_estimate)]) 
        normalized_noise = noise_size*(size_factor/len(exp))
        noise_cutoff=normalized_noise
         
        nodeDict = {}
        nodeDict[geneID] = nodes

        model_labelsDict = {}
        model_labelsDict[geneID] = label_pred
        df_list = calc_distance_com([geneID],nodeDict,model_labelsDict, locs, data_norm,tissue_mat)
        
        if noise_cutoff<6 and noise_cutoff>0:  ## 'Vstm4' have no 0<noise<5 , just skip 5.
            hamming_dist_list.append(df_list[1][0])
            jaccard_dist_list.append(df_list[2][0])
            hausdorff_dist_list.append(df_list[3][0])
            hamming_new_dist_list.append(df_list[4][0])
            p_min.append(min(p))
            smooth_factors.append(temp_factor)
             
        while noise_cutoff >0 and temp_factor<=60:            
            temp_factor = temp_factor+5
            if model == 'gmm':
                p,nodes,newLabels,_,com, label_pred = compute_single_gmm_fixed_sf(
                                locs,exp,cellGraph,gmm, temp_factor)
            else:
                p,nodes,newLabels,_,com, label_pred = compute_single_otsu_fixed_sf(
                                            locs,exp,cellGraph,temp_factor)
            num_isolate = count_isolate(locs,cellGraph, newLabels) 
            
            noise_size = sum(num_isolate[0:int(2*noise_size_estimate)]) 
            normalized_noise = noise_size*(size_factor/len(exp))
            noise_cutoff=normalized_noise
            
            nodeDict = {}
            nodeDict[geneID] = nodes

            model_labelsDict = {}
            model_labelsDict[geneID] = label_pred
            df_list = calc_distance_com([geneID],nodeDict,model_labelsDict, locs, data_norm,tissue_mat)
            
            if noise_cutoff<6 and noise_cutoff>0:
                hamming_dist_list.append(df_list[1][0])
                jaccard_dist_list.append(df_list[2][0])
                hausdorff_dist_list.append(df_list[3][0])
                hamming_new_dist_list.append(df_list[4][0])
                p_min.append(min(p))
                smooth_factors.append(temp_factor)

        if len(hausdorff_dist_list)>0:
            genes.append(geneID)
            best_p=min(p_min)  ## multiple min(p) select min(jaccard_dist) amony
            best_p_index=np.where(np.array(p_min)==best_p)[0]
            min_p_haus=[hausdorff_dist_list[i] for i in best_p_index]
            min_inde=np.where(np.array(hausdorff_dist_list)==min(min_p_haus))[0][0]
            #min_inde=np.argmin(p_min)  #jaccard_dist_list)   ## min(p) as selection standard.
            #min_inde_haus=np.argmin(hausdorff_dist_list)
            new_hamming.append(hamming_dist_list[min_inde])
            new_jaccard.append(jaccard_dist_list[min_inde])
            new_hausdorff.append(hausdorff_dist_list[min_inde])
            new_hamming2.append(hamming_new_dist_list[min_inde])
            final_sf= smooth_factors[min_inde]

            if model == 'gmm':
                p,nodes,newLabels,_,com, label_pred = compute_single_gmm_fixed_sf(
                                locs,exp,cellGraph,gmm, final_sf)
            else:
                p,nodes,newLabels,_,com, label_pred = compute_single_otsu_fixed_sf(
                                            locs,exp,cellGraph,final_sf)

            p_values.append(p)
            node_list.append(nodes)
            s_factors.append(final_sf)
            pred_labels.append(newLabels)
            model_list.append(model)
            model_labels.append(list(label_pred))
            
    best_p_values=[min(i) for i in p_values]
    labels_array = np.array(pred_labels).reshape(len(genes), pred_labels[0].shape[0])
    data_array = np.array((genes, p_values, best_p_values, s_factors, node_list, model_list , model_labels), dtype=object).T
    t_array = np.hstack((data_array, labels_array))
    c_labels = ['p_value', 'fdr',  'smooth_factor', 'nodes','model','model_labels']
    for i in np.arange(labels_array.shape[1]) + 1:
        temp_label = 'label_cell_' + str(i)
        c_labels.append(temp_label)
    result_df_new = pd.DataFrame(t_array[:,1:], index=t_array[:,0], 
                      columns=c_labels)
    
    dist_df_new = pd.DataFrame([new_hamming, new_jaccard,new_hausdorff, new_hamming2]).T        ### no norm dist
    dist_df_new.columns = ['Hamming', 'Jaccard','Hausdorff','Hamming2']
    dist_df_new.index =  genes
            
    return result_df_new, dist_df_new

## recalculate no cuts genes
def recalc_distance_com_parallel(geneList,model_list,locs, data_norm, cellGraph,
                                 tissue_mat, gmmDict):
    num_cores = mp.cpu_count()
    if num_cores > math.floor(len(geneList)/2):
         num_cores=int(math.floor(len(geneList)/2))
    ttt = np.array_split(np.array(list(geneList)),num_cores)
    tuples = [( g, m, l, d, c, t, mm) for g, m, l, d, c, t,mm in zip(
                                                        ttt, 
                                                        repeat(model_list, num_cores),
                                                        repeat(locs, num_cores),
                                                        repeat(data_norm, num_cores),
                                                        repeat(cellGraph, num_cores),
                                                        repeat(tissue_mat, num_cores),
                                                        repeat(gmmDict, num_cores))]
    dist_results = parmap.starmap(recalc_distance_com, tuples,
                             pm_processes=num_cores, pm_pbar=True)
    
    result_df_new=pd.DataFrame()
    best_dist_df=pd.DataFrame()
    for i in range(len(dist_results)):
        if dist_results[i][0].shape[0]>0 and dist_results[i][1].shape[0]>0:
            result_df_new=pd.concat([result_df_new,dist_results[i][0]])
            best_dist_df=pd.concat([best_dist_df,dist_results[i][1]])
    
    return result_df_new,best_dist_df




### ------------- Kmeans labels --------------------------
def calc_distance_kmeans(genelist,data_norm,tissue_mat, n_cluster = 5):
    ham_dist_list = []
    jac_dist_list = []
    geneList = []
    for geneID in genelist:
        X = data_norm.loc[:,geneID].values
        kmeans=KMeans(n_clusters= n_cluster,n_init=100).fit(X.reshape(-1,1))  #random_state= 0; n_init 更换初始值次数

        target_mat = []
        label,count = np.unique(kmeans.labels_, return_counts=True)
        for i in label:
            temp_vec = np.zeros(len(X))
            temp_vec[kmeans.labels_ == i]=1
            if sum(temp_vec) > len(X)/2+10:
                temp_vec = abs(1 - temp_vec)
            target_mat.append(temp_vec)

        ham_dist = cdist(tissue_mat, target_mat, compute_diff_vs_common_const10)
        ham_dist_list.append(np.min(ham_dist))

        jac_dist = cdist(tissue_mat, target_mat, jaccard_dist)
        jac_dist_list.append(np.min(jac_dist))
        geneList.append(geneID)
    
    # df = pd.DataFrame(index=geneList, 
    #         data= {'Hamming':ham_dist_list, 'Jaccard':jac_dist_list})
    return geneList, ham_dist_list, jac_dist_list

    
def calc_distance_kmeans_multi(genelist,data_norm,tissue_mat, n_cluster = 5):
    num_cores = mp.cpu_count()
    if num_cores > math.floor(len(genelist)/2):
         num_cores=int(math.floor(len(genelist)/2))
    ttt = np.array_split(np.array(list(genelist)),num_cores)
    tuples = [( g,  d,  t,  n) for g,  d,  t,  n in zip(
                                    repeat(ttt, num_cores),
                                    repeat(data_norm, num_cores),
                                    repeat(tissue_mat, num_cores),
                                    repeat(n_cluster, num_cores))]
    results = parmap.starmap(calc_distance_kmeans, tuples,
                                pm_processes=num_cores, pm_pbar=True)
    ggg = [results[i][0] for i in np.arange(len(results))]
    genes = reduce(operator.add, ggg)

    ham = [results[i][1] for i in np.arange(len(results))]
    ham = reduce(operator.add, ggg)

    jac = [results[i][2] for i in np.arange(len(results))]
    jac = reduce(operator.add, ggg)

    distance_df = pd.DataFrame(index= genes, 
            data= {'Hamming':ham, 'Jaccard':jac})

    return distance_df


    


import itertools
import operator
from functools import reduce
from scipy.spatial import distance
import numpy as np
import pandas as pd
import networkx as nx
import matplotlib as mpl
import matplotlib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.path as mplPath
from scipy.sparse import issparse
from scipy.spatial.distance import cdist
from scipy import stats  
from ast import literal_eval
import pickle
from .Graph_cut import create_graph_with_weight


def store_gmm(gmmDict,fileName):
    with open(fileName,"wb") as fw:
        pickle.dump(gmmDict,fw)

def grab_gmm(fileName):
    with open(fileName,"rb") as fr:
        return pickle.load(fr)
       

# 202111221 remove empty spots and genes by coco
def read_spatial_expression(file,sep='\s+',
                            num_exp_genes=0.01, num_exp_spots=0.01, min_expression=1,
                            drop= False):
    
    '''
    Read raw data and returns pandas data frame of spatial gene express
    and numpy ndarray for single cell location coordinates; 
    Meanwhile processing raw data.
    
    :param file: csv file for spatial gene expression; 
    :rtype: coord (spatial coordinates) shape (n, 2); data: shape (n, m); 
    '''
    counts = pd.read_csv(file, sep=sep, index_col = 0)
    print('raw data dim: {}'.format(counts.shape))
    
    # filter all empty cells/spots
    counts.fillna(0)
    counts = counts[(counts == 0).sum(axis = 1) < counts.shape[1]]

    num_spots = len(counts.index)
    num_genes = len(counts.columns)
    min_genes_spot_exp = round((counts != 0).sum(axis=1).quantile(num_exp_genes))
    print("Number of expressed genes a spot must have to be kept " \
    "({}% of total expressed genes) {}".format(num_exp_genes, min_genes_spot_exp))
    
    if counts.shape[0] > 1000:   # too many cells/spots
        drop = True
    else:
        drop = drop
        
    if drop:
        counts = counts[(counts != 0).sum(axis=1) >= min_genes_spot_exp]
        noiseInd =[]
    else:
        mark_points = np.where((counts != 0).sum(axis=1) < min_genes_spot_exp)[0]
        print("Marked {} spots".format(len(mark_points)))
        if len(mark_points)>0:
            noiseInd = [counts.shape[0]-1-i for i in range(len(mark_points))]
            temp = [val.split('x') for val in counts.index.values]
            coord = np.array([[float(a[0]), float(a[1])] for a in temp])

            similar_points=np.argsort(cdist(coord[mark_points,:],coord),axis=1)[:,1]
            for i,j in zip(mark_points,similar_points):
                counts.iloc[i,:]=counts.iloc[j,:]

            mark_counts=counts.iloc[mark_points,:]
            dropped_counts = counts.drop(counts.index[mark_points])
            counts=pd.concat([dropped_counts,mark_counts])       
        else:
            counts = counts
            noiseInd =[] 
    
    # Spots are columns and genes are rows
    counts = counts.transpose()
    # Remove noisy genes
    min_features_gene = round(len(counts.columns) * num_exp_spots) 
    print("Removing genes that are expressed in less than {} " \
    "spots with a count of at least {}".format(min_features_gene, min_expression))
    counts = counts[(counts == 0).sum(axis=1) < counts.shape[1]]
    counts = counts[(counts >= min_expression).sum(axis=1) >= min_features_gene]
    print("Dropped {} genes".format(num_genes - len(counts.index)))

    data=counts.transpose()
    
    # recheck all empty cells/spots after delect low expression genes. 20211119 coco
    data = data[(data==0).sum(axis =1) < data.shape[1]]
     
    coord = Get_coord(data)

    return coord, data, noiseInd


def Get_coord(data):
    '''
    data shape: cells x genes 
    row is spots/cells; columns is genes
    '''
    temp = [val.split('x') for val in data.index.values]
    locs = np.array([[float(a[0]),float(a[1])] for a in temp])
    return locs


def conver_exp_data(df, corinfo):
    '''
    df shape: genes x cell
    corinfo shape: cell x [x, y] 
    output genes x cell_coordinates
    '''
    coord = [str(corinfo.x[i])+'x'+str(corinfo.y[i]) for i in range(corinfo.shape[0])]
    df.columns = coord
    df.replace([-np.inf, np.inf], np.nan)  # del abnomral 
    df = df.dropna()
    df = df.T
    df = df.groupby(df.index).sum()  # sum the same points expression
    return df


def write_result_to_csv(df,fileName):

    '''
    For convenience, user should output and save scGCO result dataframe with this function.
    Meanwhile, output result with scGCO can be resued and readed across platforms 
            with **read_result_to_dataframe()** function.
     More detail can see **read_result_to_dataframe()**.
    '''
    df_=df.copy()
    df_.nodes=df.nodes.apply(conver_list)
    df_.to_csv(fileName)

def read_result_to_dataframe(fileName,sep=','):
    
    '''
    Read and use scGCO output file cross-platform .
    More detail can see **write_result_to_csv()**.
    '''
    converters={"p_value":converter,"nodes":converter, "model_labels":converter}
    df=pd.read_csv(fileName,converters=converters,index_col=0,sep=sep)
    df.nodes=df.nodes.apply(conver_array)
    return df

def conver_list(x):
    return [list(xx) for xx in x ]

def conver_array(x):
    return [np.array(xx) for xx in x] 

def converter(x):
    #define format of datetime
    return literal_eval(x)


def normalize_count_cellranger(data,Log=True):
    '''
    normalize count as in cellranger
    
    :param file: data: A dataframe of shape (m, n);
    :rtype: data shape (m, n);
    '''
    normalizing_factor = np.sum(data, axis = 1)/np.median(np.sum(data, axis = 1))
    data = pd.DataFrame(data.values/normalizing_factor[:,np.newaxis], columns=data.columns, index=data.index)
    if Log==True:
        data=log1p(data)
    else:
        data=data

    return data


def log1p(data):
    '''
    log transform normalized count data
    
    :param file: data (m, n); 
    :rtype: data (m, n);
    '''
    if not issparse(data):
        return np.log1p(data)
    else:
        return data.log1p()


## Adding points  ##20191010
def Find_line(locs,cellGraph,axis=1):

    '''
    group points by the same curve.
    axis=1 represents adding points on horizontal(row) with group_yy;
    axis=0 represents adding points on vertical(columns) with group_xx. 
    
    '''

    axis=axis
    locs_xy=locs[:,axis]
    xy_sort=sorted(locs_xy)
    group_=[]
    group_xy=[]
    
    edge_down=min(cellGraph[:,3])
    level=xy_sort[0]+edge_down

    for cc,xx in enumerate(xy_sort):

        if xx<=level:
            locs_index=np.argsort(locs_xy)[cc]
            group_.append(locs_index)

        else:
            group_xy.append(group_)
            group_=[]
            if cc<len(xy_sort)-1:
                level=xy_sort[cc+1]+edge_down
            else:
                level=xy_sort[cc]+edge_down
            if xx<=level:
                locs_index=np.argsort(locs_xy)[cc]
                group_.append(locs_index)

    group_xy.append(group_)   
    return group_xy


def Add_Points(group_xy,locs,axis=1):
    
    '''
    Add holes with missed points on the curve.
    axis=1 represents adding points on horizontal(row) with group_yy;
    axis=0 represents adding points on vertical(columns) with group_xx. 
    '''
    
    newPoints={}
    axis=axis
    ax=int(abs(1-axis))
    for k,curve in enumerate(group_xy):
        addNode=[]
        for gg in curve:
    #         plt.scatter(locs[gg,0],locs[gg,1])
            addNode.append(locs[gg,ax])  ## locs_x of group_yy[g]

        
        dist=[]   
        add_sort=sorted(addNode)
        for k in range(0,len(add_sort)-1):

            dd=add_sort[k+1]-add_sort[k]   
            dist.append(dd)

        for k in range(len(dist)):
            
            if dist[k]> 1.5*np.median(dist) and dist[k]<3*min(dist):
                #print(np.median(dist),min(dist))
                newPoint1=curve[np.argsort(addNode)[k]]
                newPoint2=curve[np.argsort(addNode)[k+1]]
                addPoints=str(newPoint1)+' '+ str(newPoint2)
              #  print(k,addPoints)
                x_=(locs[newPoint1,0],locs[newPoint2,0])
                y_=(locs[newPoint1,1],locs[newPoint2,1])
                #plt.scatter(x_,y_,c='b',s=100)
                newLocs=[reduce(operator.add,x_)/2,reduce(operator.add,y_)/2]
                newPoints[addPoints]=newLocs

    return newPoints

def Assign_exp_to_newPoints(data_norm,newPoints):
    '''
    assign expression to newPoints with their neighbors nodes' expression.
    para: locs is the coordinate;
          data is the clean dataFrame expression with normalization.
          newPoints is dictionary, whose keys is the two sides nodes of addPoints
                     and values is the coordinate of newPoints.
    '''
    
    
    ## 1st, appending newPoints to data
    data_norm_temp =data_norm.copy()
    newPoints_sort=sorted(newPoints.items(), key=lambda d: d[1])  ## sort by values and reture keys and values list. 
    for n,di in enumerate(newPoints_sort):
        key=di[0]
        values=di[1]
        index_=str(values[0])+'x'+str(values[1])
        #print(key,index_)

        node1=int(key.split(sep=' ')[0])
        node2=int(key.split(sep=' ')[1])
        node=list([node1,node2])
        node_=[v for v in node if v< data_norm.shape[0]]
        Add_exp=np.median(data_norm.iloc[node_,],axis=0)
        data_norm_temp.loc[index_]=Add_exp

    locs_new=Get_coord(data_norm_temp)
    
    ## 2nd, update expression with newPoints neighbors node
    
    exp=data_norm_temp.iloc[:,1].values
    cellGraph_new=create_graph_with_weight(locs_new,exp)

    G_add = nx.Graph()
    tempGraph = cellGraph_new.copy()
    G_add.add_nodes_from(list(set(list(tempGraph[:,0].astype(np.int32)) + list(tempGraph[:,1].astype(np.int32)))))
    G_add.add_edges_from(tempGraph[:, 0:2].astype(np.int32))  

    
    data_norm_new=data_norm.copy()
    newPoints_sort=sorted(newPoints.items(), key=lambda d: d[1])  ## sort by values and reture keys and values list.
    
    for n,di in enumerate(newPoints_sort):
        key=di[0]
        values=di[1]
        index_=str(values[0])+'x'+str(values[1])
        #print(n,key,index_)
        
        nodeInd=np.where(data_norm_temp.index==index_)[0][0]
        node=list(G_add.neighbors(nodeInd)) ## adding points index
        #print(node)
        node_=[v for v in node if v< data_norm.shape[0]]
       # print(node_)
        Add_exp=np.median(data_norm.iloc[node_,:],axis=0)
        data_norm_new.loc[index_]=Add_exp
    
    locs_new=Get_coord(data_norm_new)
    
    return locs_new,data_norm_new



def AddPoints_and_update_data(locs,data_norm,cellGraph,axis=1):
    
    '''
    Fix the holes on graph cuts by adding points.
    axis=1 represents adding points on horizontal(row) with group_yy;
    axis=0 represents adding points on vertical(columns) with group_xx. 
    
    '''
    
    ## 1. finding holes on row or columns.
    axis=axis
    group_xy=Find_line(locs,cellGraph,axis)
           
    ## 2. adding missed points
    
    newPoints=Add_Points(group_xy,locs,axis)
       
    
    ## 3. assign exprossion for newPoints
    locs_new,data_norm_new=Assign_exp_to_newPoints(locs,data_norm,newPoints)
    
    return locs_new,data_norm_new,newPoints


def AddPoints_XY_and_update_data(locs,data_norm,cellGraph,axis=1):
    '''
    Adding points as horizontal and vertical;
    return new locs, data_norm and newPoints.
    axis=1 represents adding points on horizontal(row) with group_yy;
    axis=0 represents adding points on vertical(columns) with group_xx. 
    
    '''
    newPoints={}
    locs_new_temp,data_norm_new_temp,newPoints_temp = AddPoints_and_update_data(locs,data_norm,cellGraph,axis=1)
    newPoints.update(newPoints_temp)
    
    i=1
    while len(newPoints_temp) >0:
        i=i+1
        ## 3rd, adding points on vertical with group_xx.
        exp_new=data_norm_new_temp.iloc[:,1].values
        cellGraph_new=create_graph_with_weight(locs_new_temp,exp_new)

        group_xy=Find_line(locs_new_temp,cellGraph_new,axis=int(i%2))
        newPoints_temp =Add_Points(group_xy,locs_new_temp,axis=int(i%2))
      
        
        ## 4th, update newPoints and locs_new as x_axis. 
        newPoints.update(newPoints_temp)
        locs_new_temp,data_norm_new_temp=Assign_exp_to_newPoints(locs_new_temp,data_norm,newPoints)
     
    return locs_new_temp,data_norm_new_temp,newPoints
    
def remove_points(x, noiseInd):
    x_ = x.copy()
    noiseInd.sort(reverse=True)
    for In in noiseInd:
        if len(x_)-1 >=In:
#             print(len(x_),In)
            del x_[In]
    return x_

def remove_points_in_list(node,p, noiseInd):
    node_list =[]
    zeroInd = []
    
    for n in range(len(node)):
        node_ = np.array([v for v in node[n] if v not in noiseInd])
        if len(node_)>0:
            node_list.append(list(node_))
        else:
            zeroInd.append(n)
       
    if len(zeroInd)>0:
        p_list = remove_points(p, zeroInd)
    else:
        p_list = p
    
    return node_list, p_list

def conver_str(x):
    return eval(x)

# 20201030 by coco
def update_result_df_delPoints(result_df_new,dropPoints):
    '''
    Update result_df_new with added points.
    return the same spots/cells result_df.
    '''
    
    noiseInd = dropPoints
    df = result_df_new.iloc[:,:-len(noiseInd)].copy()  # update newLabels
    for j in range(df.shape[0]):
        new_node, new_p = remove_points_in_list(df.nodes[j], df.p_value[j], noiseInd)
        df.iloc[j,0] = str(new_p)      # update p_value
        df.iloc[j,3] = str(new_node)    # update nodes
        
        new_model_labels = remove_points(df.model_labels[j], noiseInd)
        df.iloc[j,5] = str(new_model_labels)  # update 'model_labels'
    
    df.p_value = df.p_value.apply(conver_str)
    df.nodes = df.nodes.apply(conver_str)
    df.model_labels = df.model_labels.apply(conver_str)
    return df



def update_result_df_delPoints_old(result_df_new,data_norm_new,data_norm):
    '''
    deprecated and modify with update_result_df_delPoints()
    Update result_df_new with added points.
    return the same spots/cells result_df.
    
    '''
    noisePoints=[ val for val in data_norm_new.index.values if val not in data_norm.index.values]
    noiseInd=[np.where(val==data_norm_new.index.values)[0][0] for val in noisePoints if val in data_norm_new.index.values]
    print('del_points_number: ',len(noiseInd))
    
    nodes_array=[]
    newLabels_array=[]
    p_array = []
    genes=[]
    for geneID in result_df_new.index.values:  #range(result_df_new.shape[0]):
        # print(geneID)
        genes.append(geneID)
        node_=result_df_new.loc[geneID,'nodes']
        p = result_df_new.loc[geneID,'p_value']
        Labels=result_df_new.loc[geneID,][4:].values
        newLabels=[Labels[v] for v in range(len(Labels)) if v not in noiseInd]
        newLabels_array.append(newLabels)

        nodes=[]
        zeroInd = []
        for n in range(len(node_)):
            node=np.array([v for v in node_[n] if v not in noiseInd])
            if len(node)>0:
                nodes.append(node)
            else:
                zeroInd.append(n)
        nodes_array.append(nodes)
        if len(zeroInd)>0:
            zeroInd.sort(reverse=True)
            for nn in zeroInd:
                del p[nn]
        p_array.append(p)
        
    labels_array = np.array(newLabels_array).reshape(len(genes), data_norm.shape[0])
    data_array = np.array((genes,p_array, nodes_array), dtype=object).T
    t_array = np.hstack((data_array, labels_array))
    c_labels = ['p_value','nodes']
    for i in np.arange(labels_array.shape[1]) + 1:
        temp_label = 'label_cell_' + str(i)
        c_labels.append(temp_label)
    temp_df = pd.DataFrame(t_array[:,1:], index=t_array[:,0], 
                      columns=c_labels)

    result_df_delPoints = pd.concat([result_df_new.iloc[:,1:3],temp_df],axis=1)
    return result_df_delPoints